const toA = (edge) => {
  if (Array.isArray(edge)) {
    return edge
  } else {
    return [edge.child, edge.parent]
  }
}

class Digraph {
  // edges maybe either [child, parent] or { child, parent }
  constructor (edges = []) {
    this.edges = edges
  }

  /*
  edges () {
    return this.edges
  }
*/
  acdcs (s, from, to) {
    const todo = [s]
    const seen = new Set([s])
    const acdcs = new Set([])
    while (todo.length > 0) {
      const n = todo.pop()
      this.edges.forEach((e) => {
        e = toA(e)
        if (e[from] === n) {
          acdcs.add(e[to])
          if (!seen.has(e[to])) {
            todo.push(e[to])
            seen.add(e[to])
          }
        }
      })
    }
    return acdcs
  }

  isA (low, high) {
    if (low === high) {
      return true
    }
    return this.ancestors(low).has(high)
  }

  lessThan (low, high) {
    if (low === high) {
      return false
    }
    return this.ancestors(low).has(high)
  }

  descendants (s) {
    return this.acdcs(s, 1, 0)
  }

  ancestors (s) {
    return this.acdcs(s, 0, 1)
  }

  minima (nodes) {
    let minima = new Set(nodes)
    const ancestors = new Set([])
    nodes.forEach((node) => {
      this.ancestors(node).forEach((n) => ancestors.add(n))
    })
    ancestors.forEach((n) => minima.delete(n))
    if (minima.size === 0) {
      // all unrelated
      minima = new Set(nodes)
    }
    return minima
  }

  maxima (nodes) {
    const maxima = new Set(nodes)
    const descendants = new Set([])
    nodes.forEach((node) => {
      this.descendants(node).forEach((n) => descendants.add(n))
    })
    descendants.forEach((n) => maxima.delete(n))
    return maxima
  }

  add (child, parent) {
    this.edges.push([child, parent])
  }

  addList (l) {
    for (let i = 1; i < l.length; ++i) {
      this.edges.push([l[i - 1], l[i]])
    }
  }

  order (todo) {
    const ordered = []
    while (todo.length > 0) {
      const nodes = this.minima(todo)
      todo = todo.filter((e) => !nodes.has(e))
      for (const node of nodes) {
        ordered.push(node)
      }
      // ordered = ordered.concat([...nodes])
    }
    return ordered
  }
}

module.exports = Digraph
